// ExternalFunctionLibraryManager.cpp: implementation of the CExternalFunctionLibraryManager class.
//
//////////////////////////////////////////////////////////////////////

/*
   Copyright 2013 Tony Kennedy, Ventana Systems UK

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

#include "stdafx.h"

#include "ExternalFunctionLibrary.h"
#include "ExternalFunctionLibraryManager.h"

#include "MultiExtLib.h"

//#include "log.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// http://stackoverflow.com/questions/557081/how-do-i-get-the-hmodule-for-the-currently-executing-code
HMODULE GetCurrentModule()
{ 
	
  HMODULE hModule = NULL;
  GetModuleHandleEx(
    GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
    (LPCTSTR)GetCurrentModule,
    &hModule);

  return hModule;
}




void GetDLLNameFromPath(CString *pstrDLLName)
{
	long nBackSlashPos = pstrDLLName->ReverseFind('\\');
	if ( nBackSlashPos >= 0 )
	{
		*pstrDLLName = pstrDLLName->Mid(nBackSlashPos+1);

		//now remove the ".DLL"
		CString strTempString = *pstrDLLName;
		
		strTempString.MakeLower();

		nBackSlashPos = strTempString.Find(_T(".dll"));

		if ( nBackSlashPos >= 0 )
		{
			*pstrDLLName = pstrDLLName->Left(nBackSlashPos);
		}
	}
}


long RemoveFileNameFromString(CString *pString)
{
	long nBackSlashPos = pString->ReverseFind('\\');
	if ( nBackSlashPos >= 0 )
	{
		*pString = pString->Left(nBackSlashPos+1);
		return pString->GetLength();
	}
	return -1;
}


//http://stackoverflow.com/questions/6924195/get-dll-path-at-runtime
CString GetPluginPath()
{ 
	HMODULE hModule = NULL;

	hModule = GetCurrentModule();

    if (!hModule)
    {
        int ret = GetLastError();
        fprintf(stderr, "GetModuleHandle returned %d\n", ret);
    }

	TCHAR strPath[MAX_PATH];

    GetModuleFileName(hModule, strPath, MAX_PATH); // todo: should be unicode
	
	return strPath;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CExternalFunctionLibraryManager::CExternalFunctionLibraryManager(CString vensiminifile, int ext_version)
{
#ifndef NOLOG
	m_pLogFile = NULL;
#endif

	BOOL bGotNext = TRUE;
	CString strDLL; // name of dll to load
	CString strManager; // path to this plugin manager
	CString strPluginsDir; // plugin directory
	CString strPluginsDirSearchPattern; // plugin directory
	CString strLog; // log file
	CString strExtLibDLLPath = GetPluginPath(); // this lib
	// todo: test for valid path?
	WIN32_FIND_DATA FindFileData;
	HANDLE hFind;

#ifndef NOLOG
	// set up log file
	m_strLogFileName = strExtLibDLLPath;
	m_strLogFileName += _T(".log");

	OpenLogFile(FALSE);
	WriteLogEntry( _T("\nStarting ") + strExtLibDLLPath + _T(" external function library manager.") );
#endif

	// get the filename of the manager (this dll)
	strManager = strExtLibDLLPath;
	PathStripPath(strManager.GetBuffer());

	// get the search path, i.e. abs path to plugin dir (loc of manager) plus \*.dll
	strPluginsDir = strExtLibDLLPath;
	RemoveFileNameFromString(&strPluginsDir);
	strPluginsDirSearchPattern = strPluginsDir + _T("*.dll");

	hFind = FindFirstFile(strPluginsDirSearchPattern, &FindFileData);
	while ( hFind != INVALID_HANDLE_VALUE && bGotNext ) 
	{
		if( strManager.CompareNoCase(FindFileData.cFileName) != 0 )
		{
			CExternalFunctionLibrary *pLib = new CExternalFunctionLibrary(this);
			BOOL bSuccess = FALSE;

			// create an abs path to load
			//strDLL = strPluginsDir;
			//PathRemoveFileSpec(strDLL.GetBuffer());
			strDLL = strPluginsDir + FindFileData.cFileName;

			try
			{
				CString strDLLName = strDLL;
				GetDLLNameFromPath(&strDLLName);
				bSuccess = pLib->LoadDLL(strDLL, strDLLName);
			}
			catch(...)
			{
				delete pLib;
				WriteLogEntry(_T("\nFailed to load ") + strDLL) ;
			}

			if ( bSuccess )
			{
				// check version before adding - must match manager (manager will later be checked by Vensim)
				int version = pLib->_version_info();
				if ( ext_version == version ) { // success
					m_arrExtFunc.Add(pLib);
					WriteLogEntry(_T("\nLoaded ") + strDLL );
				}
				else { // mismatch - most likely single vs double precision issue
					WriteLogEntry(_T("Version mismatch - not loading " ) + pLib->GetDLLName() );
					CString strMessage;
					strMessage.Format( _T("Manager: %i, library: %i"), ext_version, version );
					WriteLogEntry( strMessage );
				}
			}
		}
		bGotNext = FindNextFile(hFind, &FindFileData);
	}
	FindClose(hFind);

	// prepare for user_definition call by pointing to lib 0
	
	//ok. So all libs now loaded and functions read in. Check for duplicates.
	FindAndRenameDuplicateFunctions();
	
	//Now set up the function IDs for Vensim
	SetUpFunctionIDs();
}


CExternalFunctionLibraryManager::~CExternalFunctionLibraryManager()
{
	CExternalFunctionLibrary *pLib;

	WriteLogEntry(_T(""));

	for ( int i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		pLib = m_arrExtFunc.GetAt(i);
		WriteLogEntry(_T("Exiting - unloading ") + pLib->GetDLLName() );
		delete pLib;
	}
	m_arrExtFunc.RemoveAll();

	CloseLogFile();
}

int CExternalFunctionLibraryManager::SetAllGV(GLOB_VARS *vgv)
{
	int nResult = 0;
	CExternalFunctionLibrary *pLib;
	
	// set gv for each lib
	for ( int i = m_arrExtFunc.GetUpperBound() ; i >=0 ; i-- ) //loop backwards in case a lib needs to be deleted.
	{
		pLib = m_arrExtFunc.GetAt(i);
		
		if ( pLib != NULL )
		{
			nResult  = pLib->_set_gv(vgv);
		
			if (!nResult )
			{	
				// failed to set gv - unload	
				(*vgv->error_message)(VERROR,(unsigned char *)"Failed set_gv - unloading external function library - see log");
				// todo: untested
				WriteLogEntry(_T("Failed set_gv - unloading " ) + m_arrExtFunc.GetAt(i)->GetDLLName() );
				pLib->UnLoadDLL();
				m_arrExtFunc.RemoveAt(i);
				
			}
		}
	}
	
	// see if any libs survived
	if ( m_arrExtFunc.IsEmpty() )
		return 0;
	else
		return 1;
}

// Check EXTERNAL_VERSION for libs - match to manager
// note: not needed; version is now checked before load
int CExternalFunctionLibraryManager::CheckAllVersionInfo(int ext_version)
{

	int version = 0;
	CExternalFunctionLibrary *pLib;
	
	// set gv for each lib
	for ( int i = m_arrExtFunc.GetUpperBound() ; i >=0 ; i-- ) //loop backwards in case a lib needs to be deleted.
	{
		pLib = m_arrExtFunc.GetAt(i);
		
		if ( pLib != NULL )
		{
		
			version = pLib->_version_info();

			if ( version != ext_version )
			{	
				// version mismatch - unload	
				// todo: pass vgv so warning can be issued to vensim error window? (*vgv->error_message)(VERROR,(unsigned char *)"Version mismatch - unloading external function library - see log");

				// todo: untested
				WriteLogEntry(_T("Version mismatch - unloading " ) + m_arrExtFunc.GetAt(i)->GetDLLName() );
				CString strMessage;
				strMessage.Format( _T("Manager: %i, library: %i"), ext_version, version );
				WriteLogEntry( strMessage );
				pLib->UnLoadDLL();
				m_arrExtFunc.RemoveAt(i);
				
			}
		}
	}
	
	// see if any libs survived
	if ( m_arrExtFunc.IsEmpty() )
		return 0;
	else
		return 1;
}


// Check funcversion_info
int CExternalFunctionLibraryManager::CheckAllFuncversionInfo(void)
{
	CString strMessage;
	WriteLogEntry(_T(""));

	// just log versions - nothing to do
	for ( int i=0; i < m_arrExtFunc.GetSize() ; i++ )
	{
		CString strMessage;
		strMessage.Format( _T(" at version: %d "), m_arrExtFunc.GetAt(i)->Get_nFuncVersion_Info() );
		WriteLogEntry(_T("Library: " ) + m_arrExtFunc.GetAt(i)->GetDLLName() + strMessage );
	}
	return 1;
}

int CExternalFunctionLibraryManager::VensimExternal(VV *val,int nval,int funcid)
{
	CExternalFunctionLibrary *pLib = GetLibByFuncID(funcid);
	
	if ( pLib == NULL )
	{
		CString strErrorMessage;
		strErrorMessage.Format(_T("Failed to locate function %d - vensim_external call fails; results may be unexpected"),funcid);
		// todo: issue a warning and cause an fp error
		WriteLogEntry(strErrorMessage);
		return 0;
	}
	
	return pLib->_vensim_external(val, nval, funcid);
}

int CExternalFunctionLibraryManager::SimulationShutdown(int finalflag)
{
	int returnval = 1;
	
	for ( int i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		int tmpretval = m_arrExtFunc.GetAt(i)->_simulation_shutdown(finalflag);
		if ( tmpretval == 0 )
		{
			WriteLogEntry(_T("Failed simulation_shutdown in ") + m_arrExtFunc.GetAt(i)->GetDLLName() );
			returnval = 0;
		}
	}
	return returnval;
}

int CExternalFunctionLibraryManager::SimulationSetup(int iniflag)
{
	int returnval = 1;
	
	for ( int i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		int tmpretval = m_arrExtFunc.GetAt(i)->_simulation_setup(iniflag);
		if ( tmpretval == 0 )
		{
			WriteLogEntry(_T("Failed simulation_setup in ") + m_arrExtFunc.GetAt(i)->GetDLLName() );
			returnval = 0;
		}
	}
	return returnval;
}

int CExternalFunctionLibraryManager::UserDefinition(int i,unsigned char **sym,unsigned char **argument_desc,int *num_arg,int *num_vector,int *func_index,int *dim_act,int *modify,int *num_loop,int *num_literal,int *num_lookup)
{

	//Find the function with i = i (hope that makes sense)
	CExternalFunctionLibrary_Function *pFunc = FindFunctionby_i(i);

	if ( pFunc != NULL )
	{
		pFunc->user_definition(sym,argument_desc,num_arg,num_vector,func_index,dim_act,modify,num_loop,num_literal,num_lookup);
		return 1;
	}

	return 0;
}

CExternalFunctionLibrary_Function* CExternalFunctionLibraryManager::FindFunctionby_i(int nIndex)
{
	long nFunctionIndex = 0;
	
	for ( long i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		CExternalFunctionLibrary *pLib_i = m_arrExtFunc.GetAt(i);

		for ( long k = 0 ; k < pLib_i->m_arr_Functions.GetSize() ; k++ )
		{
			CExternalFunctionLibrary_Function *pFunction_i = pLib_i->m_arr_Functions.GetAt(k);
			if ( pFunction_i->i == nIndex )
				return pFunction_i;
		}
	}
	return NULL;
}

void CExternalFunctionLibraryManager::SetUpFunctionIDs()
{
	long nFunctionIndex = 0;
	
	for ( long i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		CExternalFunctionLibrary *pLib_i = m_arrExtFunc.GetAt(i);

		for ( long k = 0 ; k < pLib_i->m_arr_Functions.GetSize() ; k++ )
		{
			CExternalFunctionLibrary_Function *pFunction_i = pLib_i->m_arr_Functions.GetAt(k);
			pFunction_i->SetVensimID(nFunctionIndex);
			nFunctionIndex++;
		}
	}
}

void CExternalFunctionLibraryManager::FindAndRenameDuplicateFunctions()
{
	WriteLogEntry(_T("\nChecking function names "));

	for ( long i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		CExternalFunctionLibrary *pLib_i = m_arrExtFunc.GetAt(i);

		for ( long j = 0 ; j < m_arrExtFunc.GetSize() ; j++ )
		{
			CExternalFunctionLibrary *pLib_j = m_arrExtFunc.GetAt(j);

			if ( pLib_j != pLib_i )
			{
				for ( long k = 0 ; k < pLib_i->m_arr_Functions.GetSize() ; k++ )
				{
					CExternalFunctionLibrary_Function *pFunction_i = pLib_i->m_arr_Functions.GetAt(k);
					CString strFunctionname_i = pFunction_i->sym;

					for ( long m = 0 ; m < pLib_j->m_arr_Functions.GetSize() ; m++ )
					{
						CExternalFunctionLibrary_Function *pFunction_j = pLib_j->m_arr_Functions.GetAt(m);
						CString strFunctionname_j = pFunction_j->sym;

						if ( strFunctionname_j.CompareNoCase(strFunctionname_i) == 0 )
						{
							//match found, rename both
							pFunction_i->RenameFunction( pLib_i->GetDLLName() );
							pFunction_j->RenameFunction( pLib_j->GetDLLName() );

							WriteLogEntry(_T("\tRenaming functions due to name conflict : ") 
									+ pLib_i->GetDLLName() + _T(" ") + CString(pFunction_i->sym)
									+ _T(" : ")
									+ pLib_j->GetDLLName() + _T(" ") + CString(pFunction_j->sym)
									);
						}
					}
				}
			}
		}
	}
}



CExternalFunctionLibrary* CExternalFunctionLibraryManager::GetLibByFuncID(int nFunctionID)
{
	for ( int i = 0 ; i < m_arrExtFunc.GetSize() ; i++ )
	{
		CExternalFunctionLibrary *pLib = m_arrExtFunc.GetAt(i);
		
		for ( int j = 0 ; j < pLib->m_arr_Functions.GetSize() ; j++ )
		{
			CExternalFunctionLibrary_Function *pFunc = pLib->m_arr_Functions.GetAt(j);
			
			if ( nFunctionID == pFunc->func_index_Vensim )
				return pLib;
		}
	}

	return NULL;
}


void CExternalFunctionLibraryManager::CloseLogFile()
{
#ifndef NOLOG
	if ( m_pLogFile != NULL )
	{
		m_pLogFile->Close();
		delete m_pLogFile;
		m_pLogFile = NULL;
	}
#endif
}


long CExternalFunctionLibraryManager::OpenLogFile(BOOL bAppend)
{
#ifndef NOLOG
	CloseLogFile();

	m_pLogFile = new TextFile();
	int nResult = 0;

	if ( bAppend )
	{
		m_pLogFile->Open(m_strLogFileName, TF_APPEND);
	}
	else
	{
		m_pLogFile->Open(m_strLogFileName, TF_WRITE);
	}

	if ( nResult != 0 )
	{
		//Failed to open for some reason, alert user.
		AfxMessageBox(_T("Failed to open log file. Nothing will be written!"));
		return FALSE;
	}

	if ( !bAppend )
	{
		//Log file is open for the first time, write an entry.
		WriteLogEntry(_T("Log file opened : MultiExtLib starting") );
		CloseLogFile();
	}
	else
	{
		//Close, and from now on append to the log.
		//CloseLogFile();
	}
#endif
	return TRUE;
}

void CExternalFunctionLibraryManager::WriteLogEntry(CString strLogEntry)
{
#ifndef NOLOG
	if ( m_pLogFile == NULL )
	{
		OpenLogFile(TRUE);
	}

	if ( m_pLogFile != NULL )
	{
		BOOL bDataLost = FALSE;
		m_pLogFile->WriteString(&bDataLost,strLogEntry);
		
		if ( bDataLost )
		{
			//Some error has occured.
			ASSERT(FALSE);
		}

		m_pLogFile->WriteString(&bDataLost,_T("\n"));
		
		if ( bDataLost )
		{
			//Some error has occured.
			ASSERT(FALSE);
		}
		
		CloseLogFile();

		return;
	}
#else
	return;
#endif

}



